import logging
import datetime
from typing import List, Optional, Tuple, Iterable

from . import ExpenseLog
from ._config import get_ticktick_ids, check_ticktick_ids
from ._ticktick_api import TicktickAPI
from .data.ticktick_payloads import TicktickPayloads
from .data.ticktick_id_keys import TicktickIdKeys as tik
from .data.ticktick_task_parameters import TicktickTaskParameters as ttp
from .data.ticktick_list_parameters import TicktickListParameters as tlp
from .task_model import Task
from ._task_utils import (_is_task_an_idea, _is_task_an_expense_log, _is_task_active,
                          dict_to_task, _parse_expense_log, _is_task_a_weight_measurement)

current_date = datetime.datetime.utcnow()
date_two_weeks_ago = (current_date - datetime.timedelta(days=14)).strftime("%Y-%m-%d")
date_tomorrow = (current_date + datetime.timedelta(days=1)).strftime("%Y-%m-%d")


class TicktickClient:
    """Ticktick client."""
    BASE_URL = "/api/v2"
    GET_STATE_URL = BASE_URL + "/batch/check/0"
    CRUD_TASK_URL = BASE_URL + "/batch/task"
    TASK_URL = BASE_URL + "/task"
    HABIT_CHECKINS_URL = BASE_URL + "/habitCheckins/query"
    COMPLETED_TASKS_URL = BASE_URL + f"/project/all/closed?from={date_two_weeks_ago}%2005:00:00&to={date_tomorrow}" \
                                     f"%2004:59:00&status=Completed&limit=500"
    ABANDONED_TASKS_URL = BASE_URL + f"/project/all/closed?from={date_two_weeks_ago}%2005:00:00&to={date_tomorrow}" \
                                     f"%2004:59:00&status=Abandoned&limit=500"
    DELETED_TASKS_URL = BASE_URL + "/project/all/trash/pagination?start=0&limit=500"

    def __init__(self, username: str, password: str):
        self.ticktick_client = TicktickAPI(username, password)
        self.ticktick_data: dict = {}
        self.project_lists: List[str] = []
        self._cached_raw_active_tasks: List[dict] = []
        self.all_active_tasks: List[Task] = []
        self.active_tasks: List[Task] = []
        self.completed_tasks: List[Task] = []
        self.deleted_tasks: List[Task] = []
        self.abandoned_tasks: List[Task] = []
        self.ideas: List[Task] = []
        self.weight_measurements: List[Task] = []
        self.expense_logs: List[Tuple[Task, ExpenseLog]] = []

        self._get_all_tasks()

    @staticmethod
    def _parse_ticktick_tasks(raw_tasks: List[dict] | dict, valid_ticktick_lists_ids: Optional[list] = None) \
            -> List[Task]:
        """Parses raw tasks from Ticktick into Task objects.

        Args:
            raw_tasks: Raw tasks from Ticktick.
            valid_ticktick_lists_ids: Ticktick lists ids to filter tasks by. If it is set to None, all tasks are parsed.

        Returns:
            Parsed tasks.
        """
        ticktick_tasks = []
        valid_ticktick_lists = valid_ticktick_lists_ids if valid_ticktick_lists_ids else []

        if not isinstance(raw_tasks, list):
            raw_tasks = [raw_tasks]

        if check_ticktick_ids() and get_ticktick_ids().get(tik.LIST_IDS.value, False):
            valid_ticktick_lists += get_ticktick_ids()[tik.LIST_IDS.value].values()

        for raw_task in raw_tasks:
            if valid_ticktick_lists and raw_task[ttp.PROJECT_ID.value] not in valid_ticktick_lists:
                continue

            ticktick_tasks.append(dict_to_task(raw_task))
        return ticktick_tasks

    @staticmethod
    def _get_folder_lists(lists: List[dict]) -> Optional[dict]:
        """Gets all folder lists from Ticktick.

        Args:
            lists: Raw lists from Ticktick.

        Returns:
            Project lists with this format {list_id: list_name}.
        """
        folder_lists = {}
        folder_ids: Iterable[str] = []
        if check_ticktick_ids() and get_ticktick_ids().get(tik.FOLDER_IDS.value, False):
            folder_ids = get_ticktick_ids()[tik.FOLDER_IDS.value].values()
        if folder_ids:
            folder_lists = {project_list[tlp.NAME]: project_list[tlp.ID] for project_list in lists
                            if project_list[tlp.GROUP_ID] in folder_ids}

        return folder_lists

    def _get_ticktick_data(self):
        """Gets raw data from Ticktick."""
        self.ticktick_data = self.ticktick_client.get(self.GET_STATE_URL).json()
        self.project_lists = list(self._get_folder_lists(self.ticktick_data["projectProfiles"]).values())

    def _get_all_tasks(self):
        """Gets all tasks from Ticktick."""
        self._get_ticktick_data()

        raw_active_tasks = self.ticktick_data["syncTaskBean"]["update"]
        if raw_active_tasks == self._cached_raw_active_tasks:
            return

        self._cached_raw_active_tasks = raw_active_tasks
        self.all_active_tasks = self._parse_ticktick_tasks(raw_active_tasks, self.project_lists)

        weight_measurements_id = None
        if get_ticktick_ids() is not None:
            weight_measurements_id = get_ticktick_ids()[tik.LIST_IDS.value].get("weight_measurements")

        for task in self.all_active_tasks:
            if _is_task_a_weight_measurement(task, weight_measurements_id):
                self.weight_measurements.append(task)
            elif _is_task_an_idea(task):
                self.ideas.append(task)
            elif _is_task_an_expense_log(task):
                expense_log = _parse_expense_log(task)
                if expense_log is not None:
                    self.expense_logs.append((task, expense_log))
            elif _is_task_active(task):
                self.active_tasks.append(task)
            else:
                logging.warning(f"Task {task} does not have a valid status")

    def get_ideas(self) -> List[Task]:
        """Gets all the tasks which title starts with "Idea:" from Ticktick.

        Returns:
            Ideas.
        """
        self._get_all_tasks()
        return self.ideas

    def get_expense_logs(self) -> List[Tuple[Task, ExpenseLog]]:
        """Gets all the tasks which title starts with "$" from Ticktick.

        Returns:
            Expense logs.
        """
        self._get_all_tasks()
        return self.expense_logs

    def get_active_tasks(self) -> List[Task]:
        """Gets all active tasks from Ticktick.

        Returns:
            Active tasks.
        """
        self._get_all_tasks()
        return self.active_tasks

    def get_completed_tasks(self) -> List[Task]:
        """Gets all completed tasks from Ticktick.

        Returns:
            Completed tasks.
        """
        logging.info("Getting completed tasks")

        self._get_ticktick_data()
        raw_completed_tasks = self.ticktick_client.get(self.COMPLETED_TASKS_URL).json()
        self.completed_tasks = self._parse_ticktick_tasks(raw_completed_tasks, self.project_lists)

        return self.completed_tasks

    def get_deleted_tasks(self) -> List[Task]:
        """Gets all deleted tasks from Ticktick.

        Returns:
            Deleted tasks.
        """
        self._get_ticktick_data()
        raw_deleted_tasks = self.ticktick_client.get(self.DELETED_TASKS_URL).json()["tasks"]
        self.deleted_tasks = self._parse_ticktick_tasks(raw_deleted_tasks, self.project_lists)

        return self.deleted_tasks

    def get_abandoned_tasks(self) -> List[Task]:
        """Gets all abandoned tasks from Ticktick.

        Returns:
            Abandoned tasks.
        """
        self._get_ticktick_data()
        raw_abandoned_tasks = self.ticktick_client.get(self.ABANDONED_TASKS_URL).json()
        self.abandoned_tasks = self._parse_ticktick_tasks(raw_abandoned_tasks, self.project_lists)

        return self.abandoned_tasks

    def get_task(self, task_id: str) -> Task:
        """Gets task information from Ticktick using the API.

        Args:
            task_id: Ticktick id of the task to get.

        Returns:
            Task or dictionary with the task information.
        """
        task = self.ticktick_client.get(f"{self.TASK_URL}/{task_id}", token_required=True).json()
        return dict_to_task(task)

    def complete_task(self, task: Task):
        """Completes a task in Ticktick using the API."""
        payload = TicktickPayloads.complete_task(task.ticktick_id, task.project_id)
        self.ticktick_client.post(self.CRUD_TASK_URL, data=payload, token_required=True)

    def create_task(self, task: Task, column_id: Optional[str] = None) -> str:
        """Creates a task in Ticktick using the API.

        Args:
            task: Task to create.
            column_id: Column id to create the task in. If it is set to None, the task is created in the default column.

        Returns:
            Ticktick id of the created task.
        """
        payload = TicktickPayloads.create_task(task, column_id)
        response = self.ticktick_client.post(self.CRUD_TASK_URL, payload, token_required=True).json()
        return list(response["id2etag"].keys())[0]



from attrs import define


@define
class ExpenseLog:
    """Represents a Ticktick task.

    Attributes:
        date: The title of the task, in format YYYY-MM-DD.
        expense: The amount of money spent.
        product: The product or service bought.
    """
    date: str
    expense: float
    product: str



import re
from datetime import datetime
from typing import List, Optional
from dateutil import parser, tz

from . import ExpenseLog
from .data.ticktick_task_parameters import TicktickTaskParameters as ttp
from .task_model import Task


def _clean_habit_checkins(checkins: List[dict]) -> List[str]:
    """Returns a list of formatted dates from successful habit checkins.

        Args:
            checkins (List[dict]): A list of checkin dictionaries with "checkinStamp" and "status".

        Returns:
            List[str]: A list of strings representing dates of successful checkins ('YYYY-MM-DD').
                       Returns an empty list if no successful checkins are found.

        Example:
            >>> _clean_habit_checkins([{"checkinStamp": "20230925", "status": 2}])
            ['2023-09-25']
    """
    if not checkins:
        return []

    def parse_date(date: str) -> str:
        return datetime.strptime(str(date), '%Y%m%d').strftime('%Y-%m-%d')

    return [parse_date(checkin["checkinStamp"]) for checkin in checkins if checkin["status"] == 2]


def _is_task_a_weight_measurement(task: Task, weight_measurement_id: Optional[str]) -> bool:
    """Checks if a task is a weight measurement."""
    if weight_measurement_id is None:
        return False
    return task.project_id == weight_measurement_id


def _is_task_an_idea(task: Task) -> bool:
    """Checks if a task is an idea."""
    return task.title.startswith("Idea:")


def _is_task_an_expense_log(task: Task) -> bool:
    """Checks if a task is an expense log."""
    return task.title.startswith("$")


def _is_task_active(task: Task) -> bool:
    """Checks if a task is active."""
    return task.status == 0 and task.deleted == 0


def _is_task_completed(task: Task) -> bool:
    """Checks if a task is completed."""
    return task.status == 2


def _is_task_abandoned(task: Task) -> bool:
    """Checks if a task is abandoned."""
    return task.status == -1


def _is_task_deleted(task: Task) -> bool:
    """Checks if a task is deleted."""
    return task.deleted == 1


def dict_to_task(raw_task: dict) -> Task:
    """Converts a raw task to a Task object.

    Args:
        raw_task: The raw task as dictionary.

    Returns:
        A Task object.
    """
    return Task(ticktick_id=raw_task[ttp.ID.value],
                ticktick_etag=raw_task[ttp.ETAG.value],
                created_date=_get_task_date(raw_task[ttp.TIMEZONE.value], raw_task.get(ttp.CREATED_TIME.value, None)),
                status=raw_task[ttp.STATUS.value],
                title=raw_task[ttp.TITLE.value].strip(),
                focus_time=_get_focus_time(raw_task),
                deleted=raw_task.get(ttp.DELETED.value, 0),
                tags=tuple(raw_task.get(ttp.TAGS.value, ())),
                project_id=raw_task[ttp.PROJECT_ID.value],
                timezone=raw_task[ttp.TIMEZONE.value],
                due_date=_get_task_date(raw_task[ttp.TIMEZONE.value], raw_task.get(ttp.START_DATE.value, None)),
                )


def _get_focus_time(raw_task: dict) -> float:
    """Returns the focus time of a task.

    Args:
        raw_task: The raw task from Ticktick.

    Returns:
        The focus time of the task.
    """
    focus_time = 0.0
    if ttp.FOCUS_SUMMARIES.value in raw_task:
        raw_focus_time = map(lambda summary: summary[ttp.POMO_DURATION.value] + summary[ttp.STOPWATCH_DURATION.value],
                             raw_task[ttp.FOCUS_SUMMARIES.value])
        focus_time = round(sum(raw_focus_time) / 3600, 2)
    elif ttp.FOCUS_TIME.value in raw_task:
        focus_time = float(raw_task[ttp.FOCUS_TIME.value])

    return focus_time


def _get_task_date(raw_task_timezone: str, task_start_date: str) -> str:
    """Returns the date of a task taking into account the timezone.

    Args:
        raw_task_timezone: The timezone of the task.
        task_start_date: The start date of the task.

    Returns:
        Task's date in the format YYYY-MM-DD, if the task has no start date, returns an empty string.
    """
    if not task_start_date:
        return ""

    task_timezone = tz.gettz(raw_task_timezone)
    task_raw_date = parser.parse(task_start_date)

    localized_task_date = task_raw_date.astimezone(task_timezone)
    task_date = localized_task_date.strftime("%Y-%m-%d")

    return task_date


def _parse_expense_log(raw_expense_logs: Task) -> Optional[ExpenseLog]:
    """Parses raw expense logs from Ticktick into ExpenseLog objects.

    Args:
        raw_expense_logs: Raw expense logs from Ticktick.

    Returns:
        Parsed expense logs.
    """
    expense_parse = re.search(r"\$([\d\.]+)\s+(.+)", raw_expense_logs.title)
    if not expense_parse:
        return None

    date = raw_expense_logs.due_date
    if not date:
        date = datetime.now(tz.gettz(raw_expense_logs.timezone)).strftime("%Y-%m-%d")

    return ExpenseLog(date=date,
                      expense=float(expense_parse.group(1)),
                      product=expense_parse.group(2))



from datetime import datetime
from typing import Optional

from ..task_model import Task


class TicktickPayloads:

    @staticmethod
    def get_habits_checkins(habit_list: dict, after_stamp: int) -> dict:
        return {
            "habitIds": list(habit_list.keys()),
            "afterStamp": after_stamp
        }

    @classmethod
    def complete_task(cls, task_id: str, project_id: str) -> dict:
        return {
            "update": [
                {
                    "completedUserId": 114478622,
                    "status": 2,
                    "projectId": project_id,
                    "completedTime": f"{datetime.utcnow()}+0000",
                    "id": task_id
                }
            ]
        }

    @classmethod
    def create_task(cls, task: Task, column_id: Optional[str] = None) -> dict:
        return {
            "add": [
                {
                    "startDate": task.due_date if task.due_date else None,
                    "columnId": column_id,
                    "projectId": task.project_id if task.project_id else None,
                    "title": task.title,
                    "tags": task.tags if task.tags else None,
                    "timeZone": task.timezone if task.timezone else None,
                }
            ],
            "update": [],
            "delete": [],
            "addAttachments": [],
            "updateAttachments": [],
            "deleteAttachments": []
        }



import pytest
from requests import HTTPError

from tickthon._ticktick_api import TicktickAPI


@pytest.fixture(scope="module")
def ticktick_api(ticktick_info):
    return TicktickAPI(ticktick_info["username"], ticktick_info["password"])


def test_login(ticktick_api, ticktick_info):
    token = ticktick_api.login(ticktick_info["username"], ticktick_info["password"])

    assert isinstance(token, str)
    assert token != ""
    assert token.isalnum()


def test_login_with_wrong_credentials(ticktick_api):
    with pytest.raises(HTTPError):
        ticktick_api.login("wrong-username", "wrong-password")


def test_get_request(ticktick_api):
    habits_url = ticktick_api.BASE_URL + "/habits"
    response = ticktick_api.get(habits_url)

    assert response.status_code == 200
    assert isinstance(response.json(), list)


def test_get_request_with_wrong_url(ticktick_api):
    with pytest.raises(HTTPError):
        ticktick_api.get(ticktick_api.BASE_URL + "/task/a3b7c8d9e2f1a4b6c5d4e3f2")


def test_post_request(ticktick_api):
    habit_checkins_url = ticktick_api.BASE_URL + "/habitCheckins/query"
    payload = {"habitIds": ["test-request"], "afterStamp": 20230110}

    response = ticktick_api.post(habit_checkins_url, payload)

    assert response.status_code == 200
    assert isinstance(response.json(), dict)




class TicktickListParameters:

    ID = "id"
    NAME = "name"
    IS_OWNER = "isOwner"
    COLOR = "color"
    IN_ALL = "inAll"
    SORT_ORDER = "sortOrder"
    SORT_OPTION = "sortOption"
    SORT_TYPE = "sortType"
    USER_COUNT = "userCount"
    ETAG = "etag"
    MODIFIED_TIME = "modifiedTime"
    CLOSED = "closed"
    MUTED = "muted"
    TRANSFERRED = "transferred"
    GROUP_ID = "groupId"
    VIEW_MODE = "viewMode"
    NOTIFICATION_OPTIONS = "notificationOptions"
    TEAM_ID = "teamId"
    PERMISSION = "permission"
    KIND = "kind"
    TIMELINE = "timeline"
